#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/8/18
"""
.. currentmodule:: esbems
.. moduleauthor:: Pat Daburu <pat@daburu.net>

Say something descriptive about the 'esbems' module.
"""

from .esbs import EsbMixin
from ..docstrings import model
from ..base import Base, ModelMixin


@model(label='Emergency Services Boundary (EMS)')
class EsbEms(Base, ModelMixin, EsbMixin):
    """
    This table needs a description.
    """

    __tablename__ = 'esbems'


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/4/18
"""
.. currentmodule:: meta
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This module contains metadata objects to help with inline documentation of the
model.
"""

from enum import IntFlag
from sqlalchemy import Column
from typing import Any, NamedTuple, Type, Union


COLUMN_META_ATTR = '__meta__'  #: the property that contains metadata


class Requirement(IntFlag):
    """
    This enumeration describes contracts with source data providers.
    """
    NONE = 0  #: data for the column is neither requested nor required
    REQUESTED = 1  #: data for the column is requested
    REQUIRED = 3  #: data for the column is required


class Usage(IntFlag):
    """
    This enumeration describes how data may be used.
    """
    NONE = 0  #: The data is not used.
    SEARCH = 1  #: The data is used for searching.
    DISPLAY = 2  #: The data is displayed to users.


class ColumnMeta(NamedTuple):
    """
    Metadata for table columns.
    """
    label: str  #: a human-friendly column label
    description: str = 'This field needs a description.'
    nena: str or None = None  #: the equivalent NENA field
    requirement: Requirement or int = Requirement.NONE  #: the source contract
    usage: Usage or int = Usage.NONE  #: describes how data is used
    guaranteed: bool = False  #: Is the column guaranteed to contain a value?
    calculated: bool = False  #: May the column's value be calculated?

    def get_enum(
            self,
            enum_cls: Type[Union[Requirement, Usage]]
    ) -> Requirement or Usage or None:
        """
        Get the current value of an attribute defined by an enumeration.

        :param enum_cls: the enumeration class
        :return: the value of the attribute
        """
        if enum_cls == Requirement:
            return self.requirement
        elif enum_cls == Usage:
            return self.usage
        else:
            return None

        # try:
        #     return {
        #         Requirement: lambda: self.requirement,
        #         Usage: lambda: self.usage
        #     }[enum_cls]()
        # except KeyError:
        #     return None


def column(dtype: Any, meta: ColumnMeta, *args, **kwargs) -> Column:
    """
    Create a GeoAlchemy :py:class:`Column` annotated with metadata.

    :param dtype: the GeoAlchemy column type
    :param meta: the meta data
    :return: a GeoAlchemy :py:class:`Column`
    """
    c = Column(dtype, *args, **kwargs)
    c.__dict__[COLUMN_META_ATTR] = meta
    return c





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/5/18
"""
.. currentmodule:: model
.. moduleauthor:: Pat Daburu <pat@daburu.net>

Say something descriptive about the 'model' module.
"""

import pkgutil
import gcudm.models

_SKIP_ON_LOAD = [

]  #: the names of modules that are not loaded automatically with the model


def load():
    """
    Load the data model.
    """
    package = gcudm.models
    prefix = package.__name__ + '.'
    for importer, modname, ispkg in pkgutil.walk_packages(
            package.__path__, prefix):
        if modname in _SKIP_ON_LOAD:
            continue
        else:
            _ = __import__(modname)




"""
.. currentmodule:: types
.. moduleauthor:: Pat Daburu <pat@dabru.net>

This module contains custom GeoAlchemy/SQLAlchemy types.
"""
from sqlalchemy.types import TypeDecorator, CHAR
from sqlalchemy.dialects.postgresql import UUID
import uuid


class GUID(TypeDecorator):
    """
    This is a Platform-independent GUID type that uses PostgreSQL's UUID type (
    and otherwise uses CHAR(32), storing as stringified hex values.

    .. seealso::

        http://docs.sqlalchemy.org/en/latest/core/custom_types.html#backend-agnostic-guid-type
    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/8/18
"""
.. currentmodule:: esbs
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This module provides a common mixin for all emergency service boundary
object-relational model classes.
"""

from ..meta import column, ColumnMeta, Requirement, Usage
from sqlalchemy import Column, String
from geoalchemy2 import Geometry


class EsbMixin(object):
    """
    This mixin defines common attributes of emergency service boundary
    object-relational model classes.
    """
    geom = Column(Geometry('POLYGON'))

    country = column(
        String,
        ColumnMeta(
            label='Country'
        )
    )

    state = column(
        String,
        ColumnMeta(
            label='State',
            guaranteed=True
        )
    )

    county = column(
        String,
        ColumnMeta(
            label='County'
        )
    )

    agencyID = column(
        String,
        ColumnMeta(
            label='Agency ID',
            nena='Agency_ID',
            requirement=Requirement.REQUIRED,
            guaranteed=True
        )
    )

    routeURI = column(
        String,
        ColumnMeta(
            label='Service URI',
            nena='ServiceURI',
            requirement=Requirement.REQUIRED,
            guaranteed=True
        )
    )

    serviceURN = column(
        String,
        ColumnMeta(
            label='Service URI',
            nena='ServiceURI',
            requirement=Requirement.REQUIRED,
            guaranteed=True
        )
    )

    serviceNum = column(
        String,
        ColumnMeta(
            label='Service Number',
            nena='ServiceNum',
            usage=Usage.DISPLAY
        )
    )

    vCardURI = column(
        String,
        ColumnMeta(
            label='Agency vCard URI',
            nena='AVcard_URI'
        )
    )

    displayName = column(
        String,
        ColumnMeta(
            label='Display Name',
            nena='DisplayName',
            requirement=Requirement.REQUIRED,
            #usage=Usage.DISPLAY | Usage.SEARCH
        )
    )





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/5/18
"""
.. currentmodule:: road_centerlines
.. moduleauthor:: Pat Daburu <pat@daburu.net>




"""

from ..docstrings import model
from ..meta import column, ColumnMeta, Requirement, Usage
from ..base import Base, ModelMixin
from sqlalchemy import Column, Integer, String, DateTime
from geoalchemy2 import Geometry


@model(label='Road Centerlines')
class RoadCenterline(Base, ModelMixin):
    """
    Road centerlines are the principal polyline features that represent
    road segments traversable by most emergency vehicles.
    """

    __tablename__ = 'road_centerlines'

    geom = Column(Geometry('LINESTRING'))

    srcFullNam = column(
        String,
        ColumnMeta(
            label='Source Full Name',
            requirement=Requirement.REQUESTED
        )
    )

    addRngPreL = column(
        String,
        ColumnMeta(
            label='Left Address Number Prefix',
            requirement=Requirement.REQUESTED,
            nena='AdNumPre_L',
            usage=Usage.DISPLAY

        )
    )

    addRngPreR = column(
        String,
        ColumnMeta(
            label='Right Address Number Prefix',
            requirement=Requirement.REQUESTED,
            nena='AdNumPre_R',
            usage=Usage.DISPLAY

        )
    )

    fromAddL = column(
        String,
        ColumnMeta(
            label="Left 'From' Address",
            nena='FromAddr_L',
            requirement=Requirement.REQUESTED,
            usage=Usage.DISPLAY
        )
    )

    toAddL = column(
        String,
        ColumnMeta(
            label="Left 'To' Address",
            nena='ToAddr_L',
            requirement=Requirement.REQUESTED,
            usage=Usage.DISPLAY
        )
    )

    fromAddR = column(
        String,
        ColumnMeta(
            label="Right 'From' Address",
            nena='FromAddr_R',
            requirement=Requirement.REQUESTED,
            usage=Usage.DISPLAY
        )
    )

    toAddR = column(
        String,
        ColumnMeta(
            label="Right 'To' Address",
            nena='ToAddr_R',
            requirement=Requirement.REQUESTED,
            usage=Usage.DISPLAY
        )
    )

    rngType = column(
        String,
        ColumnMeta(
            label='Ranging Type'
        )
    )

    parityL = column(
        String,
        ColumnMeta(
            label='Parity Left',
            nena='Parity_L'
        )
    )

    parityR = column(
        String,
        ColumnMeta(
            label='Parity Right',
            nena='Parity_R'
        )
    )

    preMod = column(
        String,
        ColumnMeta(
            label='Street Name Pre Modifier',
            nena='St_PreMod'
        )
    )

    preDir = column(
        String,
        ColumnMeta(
            label='Street Name Pre Directional',
            nena='St_PreDir'
        )
    )

    preType = column(
        String,
        ColumnMeta(
            label='Street Name Pre Type',
            nena="St_PreTyp"
        )
    )

    preTypeSep = column(
        String,
        ColumnMeta(
            label='Street Name Pre Type Separator',
            nena='St_PreSep'
        )
    )

    strName = column(
        String,
        ColumnMeta(
            label='Street Name',
            nena='StreetName',
            requirement=Requirement.REQUIRED
        )
    )

    postType = column(
        String,
        ColumnMeta(
            label='Street Name Post Type',
            nena='St_PosTyp'
        )
    )
    
    postDir = column(
        String,
        ColumnMeta(
            label='Street Name Post Directional',
            nena='St_PosDir'
        )
    )

    postMod = column(
        String,
        ColumnMeta(
            label='Street Name Post Modifier',
            nena='St_PosMod'
        )
    )

    lgcyPreMod = column(
        String,
        ColumnMeta(
            label='Legacy Street Name Pre Modifier',
        )
    )

    lgcyPreDir = column(
        String,
        ColumnMeta(
            label='Legacy Street Name Pre Directional',
            nena=None
        )
    )
    
    lgcyPrType = column(
        String,
        ColumnMeta(
            label='Legacy Street Name Pre Type',
            nena='LSt_PreDir'
        )
    )
    
    lgcyPrTySp = column(
        String,
        ColumnMeta(
            label='Legacy Street Name Pre Type Separator'
        )
    )

    lgcyName = column(
        String,
        ColumnMeta(
            label='Legacy Street Name',
            nena='LSt_Name'
        )
    )

    lgcyType = column(
        String,
        ColumnMeta(
            label='Legacy Street Name Post Type',
            nena='LSt_Type'
        )
    )
    
    lgcyPstDir = column(
        String,
        ColumnMeta(
            label='Legacy Street Name Post Directional',
            nena='LStPosDir'
        )
    )

    lgcyPstMod = column(
        String,
        ColumnMeta(
            label='Legacy Street Name Post Modifier'
        )
    )
    
    esnL = column(
        String,
        ColumnMeta(
            label='ESN Left',
            nena='ESN_L',
            requirement=Requirement.REQUIRED
        )
    )
    
    esnR = column(
        String,
        ColumnMeta(
            label='ESN Right',
            nena='ESN_R',
            requirement=Requirement.REQUIRED
        )
    )

    msagCommL = column(
        String,
        ColumnMeta(
            label='MSAG Community Name Left',
            nena='MSAGComm_L',
            requirement=Requirement.REQUIRED,
            usage=Usage.DISPLAY
        )
    )

    msagCommR = column(
        String,
        ColumnMeta(
            label='MSAG Community Name Right',
            nena='MSAGComm_R',
            requirement=Requirement.REQUIRED,
            usage=Usage.DISPLAY
        )
    )

    countryL = column(
        String,
        ColumnMeta(
            label='Country Left',
            nena='Country_L'
        )
    )

    countryR = column(
        String,
        ColumnMeta(
            label='Country Right',
            nena='Country_R'
        )
    )

    stateL = column(
        String,
        ColumnMeta(
            label='State Left',
            nena='State_L',
            usage=Usage.DISPLAY
        )
    )

    stateR = column(
        String,
        ColumnMeta(
            label='State Right',
            nena='State_R',
            usage=Usage.DISPLAY
        )
    )

    addCodeL = column(
        String,
        ColumnMeta(
            label='Additional Code Left',
            nena='AddCode_L'
        )
    )

    addCodeR = column(
        String,
        ColumnMeta(
            label='Additional Code Right',
            nena='AddCode_R'
        )
    )

    incMuniL = column(
        String,
        ColumnMeta(
            label='Incorporated Municipality Left',
            nena='IncMuni_L',
            usage=Usage.DISPLAY
        )
    )

    incMuniR = column(
        String,
        ColumnMeta(
            label='Incorporated Municipality Right',
            nena='IncMuni_R',
            usage=Usage.DISPLAY
        )
    )
    
    unincCommL = column(
        String,
        ColumnMeta(
            label='Unincorporated Community Left',
            nena='UnincCom_L'
        )
    )
    
    unincCommR = column(
        String,
        ColumnMeta(
            label='Unincorporated Community Right',
            nena='UnincCom_R'
        )
    )














#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/5/18
"""
.. currentmodule:: __init__.py
.. moduleauthor:: Pat Daburu <pat@daburu.net>

Say something descriptive about the '__init__.py' module.
"""


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/8/18
"""
.. currentmodule:: modes
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This module can be used to describe the mode in which the library is running.
"""

from singleton_decorator import singleton


@singleton
class Modes(object):
    """
    Represents the mode in which the library is currently being run.  If
    the module has been loaded to generate documentation, the `sphinx`
    attribute should be set to `True`.
    """
    def __init__(self):
        self.sphinx = False


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/4/18
"""
.. currentmodule:: base
.. moduleauthor:: Pat Daburu <pat@daburu.net>

The GeoAlchemy declarative base for the data model is defined in this module
along with some other helpful classes.
"""
from .modes import Modes
from .meta import column, ColumnMeta, Requirement
from .types import GUID
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import String, DateTime


class Base(object):
    """
    This is the declarative base class for SQLAlchemy and GeoAlchemy objects.
    """
    pass


# If we aren't running in "Sphinx" mode to generate documentation...
if not Modes().sphinx:
    # ...we want an actual declarative base.
    Base = declarative_base()  #: This is the model's declarative base.


class ModelMixin(object):
    """
    This is the parent class for all entity classes in the model.  It defines
    common fields.
    """
    gcUnqId = column(
        GUID,
        meta=ColumnMeta(
            label='GeoComm ID',
            guaranteed=True,
            calculated=True
        ),
        primary_key=True
    )

    srcOfData = column(
        String,
        ColumnMeta(
            label='Data Source'
        )
    )

    srcLastEd = column(
        DateTime,
        ColumnMeta(
            label='Source of Last Update'
        )
    )

    uploadAuth = column(
        String,
        ColumnMeta(
            label='Upload Authority'
        )
    )

    updateDate = column(
        DateTime,
        ColumnMeta(
            label='Last Update'
        )
    )

    effective = column(
        DateTime,
        ColumnMeta(
            label='Effective Date',
            requirement=Requirement.REQUESTED
        )
    )

    expire = column(
        DateTime,
        ColumnMeta(
            label='Expiration Date',
            requirement=Requirement.REQUESTED
        )
    )

    srcUnqId = column(
        String,
        ColumnMeta(
            label='NENA ID',
            nena='RCL_NGUID',
            requirement=Requirement.REQUESTED
        )
    )

    @classmethod
    def geometry_type(cls):
        return 'LINESTRING'  # TODO: Retrieve the geometry type.




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 4/6/18
"""
.. currentmodule:: docstrings
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This module contains utilities for setting up docstrings
"""

from .meta import ColumnMeta, COLUMN_META_ATTR, Usage, Requirement
from .modes import Modes
import inspect
from sqlalchemy import Column
import sys
from titlecase import titlecase
from typing import Any, List, Set, Tuple, Type, Union


class ModelRstFormatter(object):
    """
    This class supports a number of specialized methods for creating
    reStructuredText docstrings for classes in this project.
    """
    @staticmethod
    def simplify_docstring(s: str):
        """
        Simplify a docstring by removing leading and trailing spaces,

        :param s:
        :return:
        """
        # Make sure we're working with an actual value.
        _s = s if s is not None else ''
        # Remove leading and trailing whitespace.
        return _s.strip()

    @staticmethod
    def format_line(line: str, indent: int=1, wrap: bool=True):
        """
        Format a line of reStructuredText.

        :param line: the line to format
        :param indent: the indentation level of the formatted line
        :param wrap: Should a newline be placed at the end?
        :return: the formatted line
        """
        return '{}{}{}'.format('\t' * indent, line, '\n' if wrap else '')

    def enum2tbl(self,
                 enum_cls: Type[Union[Requirement, Usage]],
                 meta: ColumnMeta,
                 excluded: Set[Any],
                 indent: int = 1):
        # Get all of the enumerated values that aren't in the exclusion
        # set.
        vals = [v for v in enum_cls if v not in excluded]
        # Let's start off with the column specification for the table.
        colspec = f"|{'|'.join(['c'] * len(vals))}|"
        lines = [
            f'.. tabularcolumns:: {colspec}', ''
        ]
        # We're going to be formatting fixed-width text.  Let's do so with
        # three lists...
        tbl_hborders = [''] * len(vals)  # the horizontal borders
        tbl_headers = [''] * len(vals)  # the table headers
        tbl_values = [''] * len(vals)  # the values
        # Let's look at each of the values.
        for i in range(0, len(vals)):
            # We need the name.
            enum_name = vals[i].name
            # The character width of the column is the length of the name
            # plus one (1) padding space on each side.
            colwidth = (len(enum_name) + 2)
            # Now that we know the width, the border for this index can be
            # defined.
            tbl_hborders[i] = '-' * colwidth
            # Title-case the numeration name and place it in the headers
            # list at the current index.
            tbl_headers[i] = f' {titlecase(enum_name)} '
            # The yes-or-no indicator will only take up a single character,
            # but we need to pad it to maintain the fixed width.
            xo = [' '] * colwidth
            # Leaving one space on the left, put a yes-or-no indicator in
            # the column.  (We're using ASCII characters which we'll
            # replace in a moment.  For some reason, the extended characters
            # seem to pad the list with an extra space.)
            xo[1] = (
                u'Y' if meta.get_enum(enum_cls) & vals[i].value else u'N'
            )
            # Build the string.
            xos = ''.join(xo)
            # Update the string with visual symbols.
            xos = xos.replace('N', '✘')
            xos = xos.replace('Y', '✔')
            # That's the text for the values list at this index.
            tbl_values[i] = xos
        # Construct the table.
        hborder = f"+{'+'.join(tbl_hborders)}+"
        lines.append(hborder)
        lines.append(f"|{'|'.join(tbl_headers)}|")
        lines.append(hborder)
        lines.append(f"|{'|'.join(tbl_values)}|")
        lines.append(hborder)
        # Indent the entire table.
        lines = [
            self.format_line(line, indent=indent, wrap=False)
            for line in lines
        ]
        lines.append('')  # A blank line must follow the table.
        # Put it all together, and...
        rst = '\n'.join(lines)
        return rst  # ...that's that.

    def col2section(self,
                    table_name: str,
                    column_name: str,
                    meta: ColumnMeta) -> str:
        """
        Format a block of reStructuredText to represent a column.

        :param table_name: the name of the table to which the column belongs
        :param column_name: the name of the column
        :param meta: the column's meta data
        :return: a block of reStructuredText
        """
        # Start by creating an internal bookmark for the column.
        lines = [f'.. _ref_{table_name}_{column_name}:']
        # Create the name of the inline image used to represent the column.
        col_img_sub = f'img_{table_name}_{column_name}'
        # Add the image definition.
        lines.append(f'.. |{col_img_sub}| image:: _static/images/column.svg')
        lines.append(self.format_line(':width: 24px', wrap=False))
        lines.append(self.format_line(':height: 24px', wrap=False))
        lines.append('')
        # Create the heading.
        heading = f'|{col_img_sub}| **{column_name}**'
        lines.append(heading)
        lines.append('^' * len(heading))
        # Add the label.
        lines.append(self.format_line(f'**{meta.label}** - ', wrap=False))
        # Add the description.
        lines.append(self.format_line(self.simplify_docstring(meta.description)))
        # Add the table of Usage values.
        lines.append(
            self.enum2tbl(enum_cls=Usage, meta=meta, excluded={Usage.NONE},
                          indent=1))
        # Add the table of Requirement values.
        lines.append(
            self.enum2tbl(enum_cls=Requirement, meta=meta,
                          excluded={Requirement.NONE}, indent=1))

        # If the meta-data indicates there is a related NENA field...
        if meta.nena is not None:
            # ...we'll include it!
            lines.append(self.format_line(f':NENA: *{meta.nena}*'))

        # Append a blank line to separate this section from the next one.
        lines.append('')
        # Put it all together.
        rst = '\n'.join(lines)
        # Congratulations, we have a formatted reStructuredText string.
        return rst

    def cls2rst(self, cls, heading: str, preamble: str=None):
        """
        Create a docstring for a model class.

        :param cls: the class
        :param heading: the heading for the reStructuredText section
        :param preamble: everything that should preceed the generated docstring
        :return: a reStructuredText docstring
        """
        lines = ['']
        # If a preamble was supplied...
        if preamble is not None:
            # ...add it to the top.
            lines.append(preamble)
        # Figure out what we're going to call the in-line table image.
        tbl_img_sub = f'img_tbl_{cls.__name__}'
        # Define the in-line table image.
        lines.append(f'.. |{tbl_img_sub}| image:: _static/images/table.svg')
        lines.append(self.format_line(':width: 24px', wrap=False))
        lines.append(self.format_line(':height: 24px', wrap=False))
        # We need a couple of blank lines.
        lines.append('')
        lines.append('')
        # Create the heading.
        table_name_header = f'|{tbl_img_sub}| {heading}'
        lines.append('-' * len(table_name_header))
        lines.append(table_name_header)
        lines.append('-' * len(table_name_header))
        # If the class has it's own docstring...
        if cls.__doc__ is not None:
            # ...append it now.
            lines.append(cls.__doc__)
            lines.append('')
        # Now add the values.
        lines.append(self.format_line(f':Table Name: {cls.__tablename__}'))
        lines.append(self.format_line(f':Geometry Type: {cls.geometry_type()}'))
        lines.append('')  # We need a blank line.
        # We're going to go find all the members within the class hierarchy that
        # seem to be columns with metadata.
        column_members: List[Tuple[str, Column]] = []
        # Let's go through every class in the hierarchy...
        for mro in inspect.getmro(cls):
            # ...updating our list with information about all the members.
            column_members.extend(
                [
                    member for member in inspect.getmembers(mro)
                    if hasattr(member[1], COLUMN_META_ATTR)
                ]
            )
        # Eliminate duplicates.
        column_members = list(set(column_members))
        column_members.sort(key=lambda i: i[0])
        # Create the RST documentation for all the column members.
        cm_docstrings = [
            self.col2section(
                table_name=cls.__tablename__,
                column_name=cm[0],
                meta=cm[1].__meta__)
            for cm in column_members
        ]
        cm_docstring = '\n'.join(cm_docstrings)
        # Add the collected docstrings for the tables.
        lines.append(cm_docstring)
        # Put it all together...
        rst = '\n'.join(lines)
        # ...and that's a block of reStructuredText.
        return rst


def model(label: str):
    """
    Use this decorator to provide meta-data for your model class.

    :param label: the friendly label for the class
    """
    def docstring(cls):
        # We'll need a docstring formatter for this job.
        docstring_formatter = ModelRstFormatter()
        # Get the class' module.
        mod = sys.modules[cls.__module__]
        mod.__doc__ = docstring_formatter.cls2rst(
            cls=cls, heading=label, preamble=mod.__doc__)
        # Return the class.
        return cls

    def modelify(cls):
        # If the label parameter hasn't already been specified...
        if not hasattr(cls, '__label__'):
            # ...update it now.
            cls.__label__ = label
        # If we're doing a documentation run...
        if Modes().sphinx:
            # ...update the docstrings.
            docstring(cls)
        # return the original class to the caller.
        return cls

    # Return the inner function.
    return modelify

